/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 *  USBSink - addfilesdialog.cc
 *
 *  Copyright (C) 2006-2007 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "addfilesdialog.h"
#include "globalstrings.h"
#include "utils.h"

#include <algorithm>
#include <functional>
#include <list>
#include <vector>

#include <glib/gerror.h>
#include <glib/gi18n.h>
#include <glib/gmessages.h>

#include <glibmm/date.h>
#include <glibmm/refptr.h>

#include <libglademm/xml.h>

#include <sigc++/sigc++.h>

#include <libgnomevfsmm/exception.h>
#include <libgnomevfsmm/uri.h>

using std::list;
using std::vector;

namespace USBSink
{

namespace UI
{

const int vpaned_position_ = 200;

typedef vector<Glib::ustring> ustring_vector;
typedef vector<Glib::ustring>::iterator ustring_vector_iter;

typedef list< Glib::RefPtr<Gnome::Vfs::Uri> > uri_list;
//typedef list< Glib::RefPtr<Gnome::Vfs::Uri> >::iterator uri_list_iter;

typedef Gtk::TreeModel::Children tree_children;

AddFilesDialog::AddFilesDialog(Gtk::Window& parent,
			       bool from_removable,
			       const Glib::ustring& removable_mount_point)
  :
  filechooser_        (0),
  toolbar_            (0),
  toolbutton_add_     (0),
  toolbutton_remove_  (0),
  toolbutton_clear_   (0),
  treeview_file_list_ (0)
{
  load_xml();
  
  dialog_->set_transient_for(parent);
  dialog_->set_modal();
  
  connect_signals();
  
  setup_filechooser(from_removable, removable_mount_point);
  vpaned_->set_position(vpaned_position_);
  setup_file_list_layout();
  
  // set drag target
  dnd_targets_.push_back(Gtk::TargetEntry("text/uri-list"));
  treeview_file_list_->drag_dest_set(dnd_targets_);
}
  
AddFilesDialog::~AddFilesDialog()
{}

void
AddFilesDialog::load_xml()
{
  using Gnome::Glade::Xml;
  
  try
  {
    const Glib::RefPtr<Xml> xml = Xml::create(glade_addfilesdialog_filename);
    
    Gtk::Dialog* dialog_ptr = 0;
    dialog_.reset(xml->get_widget("addfilesdialog", dialog_ptr));
    
    xml->get_widget("filechooser", filechooser_);
    xml->get_widget("vpaned", vpaned_);
    xml->get_widget("toolbar", toolbar_);
    xml->get_widget("toolbutton_add", toolbutton_add_);
    xml->get_widget("toolbutton_remove", toolbutton_remove_);
    xml->get_widget("toolbutton_clear", toolbutton_clear_);
    xml->get_widget("treeview_file_list", treeview_file_list_);
  }
  catch (const Gnome::Glade::XmlError& xml_err)
  {
    g_error("%s", xml_err.what().c_str());
  }
}

void
AddFilesDialog::connect_signals()
{
  dialog_->signal_response().connect(
    sigc::mem_fun(*this, &AddFilesDialog::on_response));

  toolbutton_add_->signal_clicked().connect(
    sigc::mem_fun(*this, &AddFilesDialog::on_toolbutton_add));
  
  toolbutton_remove_->signal_clicked().connect(
    sigc::mem_fun(*this, &AddFilesDialog::on_toolbutton_remove));
  
  toolbutton_clear_->signal_clicked().connect(
    sigc::mem_fun(*this, &AddFilesDialog::on_toolbutton_clear));

  treeview_file_list_->signal_drag_data_received().connect(
    sigc::mem_fun(*this, &AddFilesDialog::on_treeview_drag_data_received));
}

void
AddFilesDialog::setup_filechooser(bool from_removable,
                                  const Glib::ustring& removable_mount_point)
{
  // allow selection of multiple files
  filechooser_->set_select_multiple();

  if (from_removable)
  {
    filechooser_->set_current_folder(removable_mount_point);
  }
  else
  {
    // allow selection of non-local files
    filechooser_->set_local_only(false);
  }

  //TODO: filter file extensions
}

void
AddFilesDialog::setup_file_list_layout()
{
  // create the tree model
  treemodel_files_ = Gtk::ListStore::create(columns_files_);
  treeview_file_list_->set_model(treemodel_files_);

  // set up the treeview selection object
  file_list_selection_ = treeview_file_list_->get_selection();
  file_list_selection_->set_mode(Gtk::SELECTION_MULTIPLE);

  // append column headers
  treeview_file_list_->append_column(_("Name"), columns_files_.col_name);
  treeview_file_list_->append_column(_("Modified"), columns_files_.col_modified);
  treeview_file_list_->append_column(_("MIME type"),columns_files_.col_mime_type);
  treeview_file_list_->append_column(_("Full path"),columns_files_.col_full_path);
}

bool
AddFilesDialog::file_is_in_list(const Glib::ustring& uri)
{
  tree_children children = treemodel_files_->children();
  
  tree_children::iterator it(children.begin());
  tree_children::iterator end(children.end());

  for (; it != end; ++it)
  {
    Glib::ustring row_full_path((*it)[columns_files_.col_full_path]);
    if (row_full_path == uri)
    {
      return true;
    }
  }
  return false;
}

void
AddFilesDialog::add_uri_to_treeview(const Glib::RefPtr< Gnome::Vfs::Uri >& uri)
{
  if (uri->uri_exists())
  {
    if (! file_is_in_list(uri->get_path()))
    {
      Glib::RefPtr<Gnome::Vfs::FileInfo> fi =
	uri->get_file_info(Gnome::Vfs::FILE_INFO_GET_MIME_TYPE);
				
      Gtk::TreeModel::Row row = *(treemodel_files_->append());
			
      row[columns_files_.col_name] = uri->extract_short_name();
			
      row[columns_files_.col_modified] =
	Util::Time::get_modification_date_string(fi->get_modification_time());
			
      row[columns_files_.col_mime_type] = fi->get_mime_type();
			
      if (uri->is_local())
      {
	row[columns_files_.col_full_path] = uri->get_path();
      }
      else
      {
	// for instance, it would be smb:://hostname/dir/file
	row[columns_files_.col_full_path] = uri->to_string();
      }
    }
  }
}

void
AddFilesDialog::on_toolbutton_add()
{
  ustring_vector uris = filechooser_->get_uris();
  
  if (! uris.empty())
  {
    ustring_vector_iter end(uris.end());
    for (ustring_vector_iter it(uris.begin()); it != end; ++it)
    {
      try
      {
	Glib::RefPtr<Gnome::Vfs::Uri> uri = Gnome::Vfs::Uri::create(*it);
	add_uri_to_treeview(uri);
      }
      catch (const Gnome::Vfs::exception& ex)
      {
	g_warning("Error while getting file info for path %s: %s",
		  it->c_str(),
		  ex.what().c_str());
      }
    }
  }
}

void
AddFilesDialog::on_toolbutton_remove()
{
  typedef vector<Gtk::TreeModel::Path> path_vector;
  typedef vector<Gtk::TreeIter> tree_iter_vector;

  path_vector sel_rows = file_list_selection_->get_selected_rows();
  tree_iter_vector tree_iters;
	
  path_vector::iterator p_it(sel_rows.begin());
  path_vector::iterator p_end(sel_rows.end());

  for (; p_it != p_end; ++p_it)
  {
    tree_iters.push_back(treemodel_files_->get_iter(*p_it));
  }

  tree_iter_vector::iterator t_it(tree_iters.begin());
  tree_iter_vector::iterator t_end(tree_iters.end());

  for (; t_it != t_end; ++t_it)
  {
    treemodel_files_->erase(*t_it);
  }
}

void
AddFilesDialog::on_toolbutton_clear()
{
  treemodel_files_->clear();
}

void
AddFilesDialog::on_response(int response_id)
{
  if (response_id == Gtk::RESPONSE_OK)
  {
    tree_children children = treemodel_files_->children();
    uri_list uris(children.size());
		
    tree_children::iterator it(children.begin());
    tree_children::iterator end(children.end());
		
    for (; it != end; ++it)
    {
      try
      {
	Glib::RefPtr<Gnome::Vfs::Uri> uri =
	  Gnome::Vfs::Uri::create((*it)[columns_files_.col_full_path]);
				
	if (uri->uri_exists())
	{
	  uris.push_back(uri);
	}
      }
      catch (const Gnome::Vfs::exception& ex)
      {
	Glib::ustring path((*it)[columns_files_.col_full_path]);
	g_warning("Error while getting file info for path %s: %s",
		  path.c_str(),
		  ex.what().c_str());
      }
    }
    signal_file_uris_selected.emit(uris);
  }
  dialog_->hide();
}

void
AddFilesDialog::on_treeview_drag_data_received(
                  const Glib::RefPtr<Gdk::DragContext>& /* context */,
                  int /* x */, int /* y */,
                  const Gtk::SelectionData& selection_data,
                  guint /* info */, guint /* time */)
{
  if (selection_data.get_data())
  {
    ustring_vector targets = selection_data.get_uris();

    if (! targets.empty())
    {
      for (ustring_vector_iter it(targets.begin()); it != targets.end(); ++it)
      {
	try
	{
	  Glib::RefPtr<Gnome::Vfs::Uri> uri =
	    Gnome::Vfs::Uri::create(*it);
	  
	  if (uri->uri_exists())
	  {
	    add_uri_to_treeview(uri);
	  }
	}
	catch (const Gnome::Vfs::exception& ex)
	{
	  g_warning("Error while getting file info for path %s: %s",
		    it->c_str(),
		    ex.what().c_str());
	}
      }
    }
  }
}

} // namespace UI

} // namespace USBSink
