/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 *  USBSink - filelisteditor.cc
 *
 *  Copyright (C) 2006-2007 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "blackfilelist.h"
#include "globalstrings.h"
#include "filelist.h"
#include "filelisteditor.h"
#include "filelistloader.h"
#include "filelistregistry.h"
#include "utils.h"
#include "whitefilelist.h"

#include <list>
#include <tr1/memory>

#include <glib/gi18n.h>

#include <pcrecpp.h>

#include <libglademm/xml.h>

namespace USBSink
{

namespace UI
{

using std::list;
using std::map;

// TODO: why this fails but the iters typedef pass ?
// typedef map<Glib::ustring, FileListEditor::ListData*> list_data_map;
typedef map<Glib::ustring, FileListEditor::ListData*>::iterator list_data_iter;

typedef map<Glib::ustring, FileListEditor::ListData*>::const_iterator
  list_data_const_iter;

void
FileListEditor::init(Gtk::Window& parent)
{
  load_xml();

  dialog_->set_transient_for(parent);

  setup_lists();

  connect_signals();
} 

FileListEditor::FileListEditor(Gtk::Window& parent)
  :
  treeview_lists_          (0),
  button_new_list_         (0),
  entry_regexp_            (0),
  entry_text_to_match_     (0),
  button_test_             (0),
  label_match_info_        (0),
  button_add_              (0),
  combobox_list_type_      (0),
  textview_list_content_   (0),
  button_cancel_           (0),
  button_ok_               (0)
{
  init(parent);
}

FileListEditor::FileListEditor(Gtk::Window& parent,
                               const Glib::ustring& new_list_name)
  :
  treeview_lists_          (0),
  button_new_list_         (0),
  entry_regexp_            (0),
  entry_text_to_match_     (0),
  button_test_             (0),
  label_match_info_        (0),
  button_add_              (0),
  combobox_list_type_      (0),
  textview_list_content_   (0),
  button_cancel_           (0),
  button_ok_               (0)
{
  init(parent);

  bool status = new_list(new_list_name);

  if (! status)
  {
    g_warning("FileListEditor couldn't create requested file list '%s'",
              new_list_name.c_str());
  }
}

FileListEditor::~FileListEditor()
{
  list_data_iter it(lists_.begin());
  list_data_iter end(lists_.end());

  for ( ; it != end; ++it)
  {
    delete it->second;
  }
}

void
FileListEditor::load_xml()
{
  using Gnome::Glade::Xml;

  try
  {
    const Glib::RefPtr<Xml> xml = Xml::create(glade_filelisteditor_filename);

    Gtk::Dialog* dialog_ptr = 0;
    dialog_.reset(xml->get_widget("file_list_editor", dialog_ptr));

    xml->get_widget("treeview_lists", treeview_lists_);
    xml->get_widget("button_new_list", button_new_list_);
    xml->get_widget("entry_regexp", entry_regexp_);
    xml->get_widget("entry_text_to_match", entry_text_to_match_);
    xml->get_widget("button_test", button_test_);
    xml->get_widget("label_match_info", label_match_info_);
    xml->get_widget("button_add", button_add_);
    xml->get_widget("combobox_list_type", combobox_list_type_);
    xml->get_widget("textview_list_content", textview_list_content_);
    xml->get_widget("button_cancel", button_cancel_);
    xml->get_widget("button_ok", button_ok_);
  }
  catch (const Gnome::Glade::XmlError& xml_err)
  {
    g_error("%s", xml_err.what().c_str());
  }
}

void
FileListEditor::connect_signals()
{
  button_new_list_->signal_clicked().connect(
    sigc::mem_fun(*this, &FileListEditor::on_button_new_list_clicked));

  button_test_->signal_clicked().connect(
    sigc::mem_fun(*this, &FileListEditor::on_button_test_clicked));

  button_add_->signal_clicked().connect(
    sigc::mem_fun(*this, &FileListEditor::on_button_add_clicked));

  combobox_list_type_->signal_changed().connect(
    sigc::mem_fun(*this,
                  &FileListEditor::on_combo_list_type_selection_changed));

  button_cancel_->signal_clicked().connect(
    sigc::mem_fun(*this, &FileListEditor::on_button_cancel_clicked));

  button_ok_->signal_clicked().connect(
    sigc::mem_fun(*this, &FileListEditor::on_button_ok_clicked));
}

/**
 * Sets up the Gtk::TreeView and loads it with lists from
 * TaskHandling::FileListRegistry.
 * Each list has and associated Gtk::TextBuffer.
 */
void
FileListEditor::setup_lists()
{
  // create the tree model
  treemodel_lists_ = Gtk::ListStore::create(columns_lists_);
  treeview_lists_->set_model(treemodel_lists_);

  // set up the treeview selection object
  tree_selection_lists_ = treeview_lists_->get_selection();
  tree_selection_lists_->set_mode(Gtk::SELECTION_SINGLE);

  // append column headers
  treeview_lists_->append_column(_("Name"), columns_lists_.col_name);
  
  // load the map with all FileLists and a TextBuffer for each one
  using namespace TaskHandling;

  FileListRegistry* registry;
  list<SharedFileList> lists;

  registry = FileListRegistry::instance();
  registry->get_file_lists(lists);

  list<SharedFileList>::iterator it(lists.begin());
  list<SharedFileList>::iterator end(lists.end());

  for ( ; it != end; ++it)
  {
    ListData* data;
    Glib::RefPtr<Gtk::TextBuffer> buffer;

    data = new ListData;
    data->type_string = (*it)->get_type_string();

    buffer = Gtk::TextBuffer::create();
    buffer->set_text((*it)->get_text_string_for_display());

    data->buffer = buffer;
    data->type_changed = false;
    data->is_new = false;

    lists_[(*it)->get_name()] = data;

    // add a new row to the treeview
    Gtk::TreeModel::Row row = *(treemodel_lists_->append());

    row[columns_lists_.col_name] = (*it)->get_name();
  }

  // connect the selection function
  tree_selection_lists_->set_select_function(
    sigc::mem_fun(*this, &FileListEditor::on_treeview_select_function));

  // and select the first list
  if (! treemodel_lists_->children().empty())
  {
    tree_selection_lists_->select(treemodel_lists_->children().begin());
  }
}

bool
FileListEditor::list_name_in_use(const Glib::ustring& name) const
{
  bool result;

  list_data_const_iter it = lists_.find(name);

  if (it == lists_.end())
    result = false;
  else
    result = true;

  return result;
}

bool
FileListEditor::new_list(const Glib::ustring& list_name)
{
  bool result;

  if (list_name_in_use(list_name))
  {
    result = false;
  }
  else
  {
    // create a new internal item
    ListData* data = new ListData;

    // make black the default
    data->type_string = TaskHandling::black_file_list_type_string;
    data->buffer = Gtk::TextBuffer::create();
    data->type_changed = false;
    data->is_new = true;

    lists_[list_name] = data;

    // add a new row to the treeview
    Gtk::TreeModel::Row row = *(treemodel_lists_->append());

    row[columns_lists_.col_name] = list_name;
    tree_selection_lists_->select(row);

    // remember to set the combobox to black (default)
    combobox_list_type_->set_active(0);

    result = true;
  }

  return result;
}

bool
FileListEditor::on_treeview_select_function(
                  const Glib::RefPtr<Gtk::TreeModel>& /* model */,
                  const Gtk::TreeModel::Path& path,
                  bool path_currently_selected)
{
  Gtk::TreeModel::Row row = *(treemodel_lists_->get_iter(path));

  if ((row[columns_lists_.col_name] != selected_list_name_) &&
      (! path_currently_selected))
  {
    using namespace TaskHandling;

    // selection changed
    selected_list_name_ = row[columns_lists_.col_name];

    // set the type combobox
    list_data_iter iter = lists_.find(selected_list_name_);
    ListData* data = iter->second;

    if (data->type_string == TaskHandling::black_file_list_type_string)
    {
      combobox_list_type_->set_active(0);
    }
    else if (data->type_string == TaskHandling::white_file_list_type_string)
    {
      combobox_list_type_->set_active(1);
    }
    else
    {
      g_warning("Unknown type of list %s", selected_list_name_.c_str());
      combobox_list_type_->set_active(0);
    }

    // change text buffer to display
    textview_list_content_->set_buffer(data->buffer);
  }

  return true;
}

void
FileListEditor::on_button_new_list_clicked()
{
  bool exit = false;

  while (! exit)
  {
    // show a simple input dialog
    Gtk::Dialog dialog(_("New list"), *dynamic_cast<Gtk::Window*>(get_dialog()));

    Gtk::VBox* vbox = dialog.get_vbox();

    Gtk::HBox hbox;
    vbox->pack_start(hbox);

    Gtk::Label label(_("Enter name for a new list:"));
    Gtk::Entry entry;
    entry.set_activates_default();

    hbox.pack_start(label);
    hbox.pack_start(entry);

    Gtk::StockID stock_cancel(Gtk::Stock::CANCEL);
    Gtk::StockID stock_ok(Gtk::Stock::OK);

    dialog.add_button(stock_cancel, Gtk::RESPONSE_CANCEL);
    dialog.add_button(stock_ok, Gtk::RESPONSE_OK);

    dialog.set_default_response(Gtk::RESPONSE_OK);

    dialog.show_all_children();
  
    int result = dialog.run();
  
    if (result == Gtk::RESPONSE_OK)
    {
      const Glib::ustring list_name = entry.get_text();

      bool creation_result = new_list(list_name);

      if (! creation_result)
      {
        Gtk::MessageDialog msg_dialog(
          *dynamic_cast<Gtk::Window*>(get_dialog()),
          _("Duplicate list."),
          false,
          Gtk::MESSAGE_ERROR,
          Gtk::BUTTONS_OK);

        msg_dialog.set_title(
          Util::i18n_printf(_("Name '%s' is already in use."),
                            list_name.c_str()));
        msg_dialog.set_default_response(Gtk::RESPONSE_OK);
        msg_dialog.set_secondary_text(_("Please choose another name."));

        msg_dialog.run();

        exit = false;
      }
      else
      {
        exit = true;
      }
    }
    else
    {
      exit = true;
    }
  }
}

void
FileListEditor::on_button_test_clicked()
{
  pcrecpp::RE regexp(entry_regexp_->get_text().c_str(), pcrecpp::UTF8());

  last_regexp_ = entry_regexp_->get_text();

  if (regexp.FullMatch(entry_text_to_match_->get_text().c_str()))
  {
    label_match_info_->set_text(
      Util::i18n_printf(_("%s matches %s"),
                        entry_text_to_match_->get_text().c_str(),
                        entry_regexp_->get_text().c_str()));
  }
  else
  {
    label_match_info_->set_text(
      Util::i18n_printf(_("%s doesn't match %s"),
                       entry_text_to_match_->get_text().c_str(),
                       entry_regexp_->get_text().c_str()));
  }
}

void
FileListEditor::on_button_add_clicked()
{
  if (last_regexp_.size())
  {
    Glib::RefPtr<Gtk::TextBuffer> current_buffer =
      textview_list_content_->property_buffer();
    current_buffer->insert(current_buffer->end(), last_regexp_ + "\n");
  }
}

void
FileListEditor::on_combo_list_type_selection_changed()
{
  Gtk::TreeModel::iterator selected_row(tree_selection_lists_->get_selected());
  list_data_iter list_iter = lists_.find(selected_list_name_);

  if ((selected_row == treemodel_lists_->children().end()) ||
      list_iter == lists_.end())
  {
    // nothing to do
    return;
  }

  ListData* data = list_iter->second;

  if ((data->type_string == TaskHandling::black_file_list_type_string) &&
      combobox_list_type_->get_active_row_number() == 1)
  {
    data->type_string = TaskHandling::white_file_list_type_string;
    data->type_changed = true;
  }
  else if ((data->type_string == TaskHandling::white_file_list_type_string) &&
      combobox_list_type_->get_active_row_number() == 0)
  {
    data->type_string = TaskHandling::black_file_list_type_string;
    data->type_changed = true;
  }
}

void
FileListEditor::on_button_cancel_clicked()
{
  dialog_->hide();
}

void
FileListEditor::on_button_ok_clicked()
{
  list_data_iter it = lists_.begin();
  list_data_iter end = lists_.end();

  for ( ; it != end; ++it)
  {
    ListData* data = it->second;
    TaskHandling::FileListRegistry* registry =
      TaskHandling::FileListRegistry::instance();

    if (data->is_new)
    {
      TaskHandling::file_list_factory_from_buffer factory;

      if (data->type_string == TaskHandling::black_file_list_type_string)
      {
        factory = &TaskHandling::BlackFileList::create;
      }
      else
      {
        factory = &TaskHandling::WhiteFileList::create;
      }

      try
      {
        registry->register_new_file_list(it->first, data->buffer, factory);
      }
      catch (TaskHandling::FileListException& ex)
      {
        g_warning("Couldn't register file list '%s': %s (unexpected)",
                  it->first.c_str(), ex.what().c_str());
      }

      continue;
    }

    if (data->type_changed)
    {
      TaskHandling::file_list_factory_from_buffer factory;

      if (data->type_string == TaskHandling::black_file_list_type_string)
      {
        factory = &TaskHandling::BlackFileList::create;
      }
      else
      {
        factory = &TaskHandling::WhiteFileList::create;
      }

      registry->reset_file_list(it->first, data->buffer, factory);
    }
    else // just reset contents
    {
      list<TaskHandling::SharedItem> new_items;

      TaskHandling::SharedFileList list =
        std::tr1::dynamic_pointer_cast<TaskHandling::FileList>(
          registry->get_file_list(it->first));

      TaskHandling::FileListLoader::get_items(data->buffer, new_items);
      list->set_items(new_items);
    }
  }

  dialog_->hide();
}

} // namespace UI

} // namespace USBSink
