/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 *  USBSink - filelistregistry.cc
 *
 *  Copyright (C) 2006-2007 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "common/nmv-log-stream-utils.h"
#include "blackfilelist.h"
#include "clientdispatcher.h"
#include "filelistregistry.h"
#include "utils.h"
#include "waitingclient.h"
#include "whitefilelist.h"

#include <vector>

#include <glib/gstdio.h>

using std::auto_ptr;
using std::list;
using std::map;
using std::vector;

namespace USBSink
{

namespace TaskHandling
{

typedef std::map<Glib::ustring, SharedFileList> file_list_map;
typedef file_list_map::iterator file_list_iter;
typedef file_list_map::const_iterator file_list_const_iter;


static const Glib::ustring file_list_not_found_msg = "File list not found.";
static const Glib::ustring file_list_exists_msg = "File list already exists.";

bool FileListRegistry::initialized_ = false;
auto_ptr<FileListRegistry> FileListRegistry::instance_;

/// Initializes internal record keeping.
FileListRegistry::FileListRegistry()
{
  load_task_file_lists();

  //
  vector<SharedWaitingClient> waiting_clients;
  vector<Glib::ustring> client_requests;

  ClientDispatcher::instance()->get_clients(waiting_clients);

  vector<SharedWaitingClient>::iterator client_it(waiting_clients.begin());
  vector<SharedWaitingClient>::iterator client_end(waiting_clients.end());

  for ( ; client_it != client_end; ++client_it)
  {
    (*client_it)->get_requested_list_names(client_requests);

    vector<Glib::ustring>::iterator req_it(client_requests.begin());
    vector<Glib::ustring>::iterator req_end(client_requests.end());

    for ( ; req_it != req_end; ++req_it)
    {
      try
      {
        SharedItem item = get_file_list(*req_it);

        ClientSlot slot = (*client_it)->get_slot();

        slot(item);
      }
      catch (FileListNotFoundException& )
      {
        // ignore
      }
    }

    client_requests.clear(); // before starting over with a new client
  }
}

FileListRegistry::~FileListRegistry()
{
}

FileListRegistry*
FileListRegistry::instance()
{
  if (! instance_.get())
    instance_.reset(new FileListRegistry());

  initialized_ = true;

  return instance_.get();
}

/** Reads all files from file_list_dir and creates shared objects to be used
 * by the rest of the application.
 */
void
FileListRegistry::load_task_file_lists()
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  try
  {
    list<std::string> lists =
      Util::File::get_files_in_dir(Util::Config::get_file_list_dir());

    list<std::string>::iterator it(lists.begin());
    list<std::string>::iterator end(lists.end());

    Glib::ustring line, list_name;

    for ( ; it != end; ++it)
    {
      line = Util::File::peek_first_line(*it);
      
      if (line == TaskHandling::black_file_list_type_string)
      {
	SharedFileList sfl = BlackFileList::create_from_file(*it);

	list_name = Glib::path_get_basename(sfl->get_path());
	lists_[list_name] = sfl;

	LOG_DD("Loaded black list: " << list_name);
      }
      else if (line == TaskHandling::white_file_list_type_string)
      {
	SharedFileList sfl = WhiteFileList::create_from_file(*it);

	list_name = Glib::path_get_basename(sfl->get_path());
	lists_[list_name] = sfl;

	LOG_DD("Loaded white list: " << list_name);
      }
    }
  }
  catch (Glib::FileError& ex)
  {
    g_critical("Error while loading task file lists: %s", ex.what().c_str());
    //TODO: propagate to UI
  }
}

/**
 * @param name Non-formatted name used to identify a FileList.
 *
 * @returns A corresponding SharedItem to use.
 *
 * @throws FileListNotFoundException
 */
SharedItem
FileListRegistry::get_file_list(const Glib::ustring& name)
{
  file_list_iter it = lists_.find(name);

  if (it == lists_.end())
    throw FileListNotFoundException(file_list_not_found_msg);
  else
    return it->second;
}

/**
 * @param ret_list A list for returning all known FileLists in use.
 */
void
FileListRegistry::get_file_lists(std::list<SharedFileList>& ret_list)
{
  file_list_iter it = lists_.begin();
  file_list_iter end = lists_.end();

  for ( ; it != end; ++it)
  {
    ret_list.push_back(it->second);
  }
}

/**
 * @param name Non-formatted name used to identify a new FileList.
 * @param factory Pointer to function used to create the file list,
 * eg &BlackFileList::create.
 *
 * @returns A SharedItem to use.
 *
 * @throws FileListExistsException
 */
SharedItem
FileListRegistry::register_new_file_list(const Glib::ustring& name,
                                         file_list_factory factory)
{
  file_list_iter it = lists_.find(name);

  if (it != lists_.end())
  {
    throw FileListExistsException(file_list_exists_msg);
  }
  else
  {
    SharedFileList sfl = factory(name);

    lists_[name] = sfl;

    return sfl;
  }
}

/**
 * @param name Non-formatted name used to identify a new FileList.
 * @param text_buffer Text buffer containing other file list names
 * and regular expressions.
 * @param factory Pointer to function used to create the file list from
 * a Gtk::TextBuffer, eg &BlackFileList::create.
 *
 * @returns A SharedItem to use.
 *
 * @throws FileListException
 */
SharedItem
FileListRegistry::register_new_file_list(
  const Glib::ustring& name,
  const Glib::RefPtr<Gtk::TextBuffer>& text_buffer,
  file_list_factory_from_buffer factory)
{
  file_list_iter it = lists_.find(name);

  if (it != lists_.end())
  {
    throw FileListExistsException(file_list_exists_msg);
  }
  else
  {
    try
    {
      SharedFileList list = factory(name, text_buffer);

      lists_[name] = list;

      return list;
    }
    catch (FileListException& ex)
    {
      throw ex; // re-throw
    }
  }
}

/**
 * @param name Non-formatted name used to identifying a FileList.
 * @param buf Text buffer containing other file list names
 * and regular expressions.
 * @param factory Pointer to function used to create the file list from
 * a Gtk::TextBuffer, eg &BlackFileList::create.
 *
 * Emits signal file_list_reset on the new list.
 * Use get_signal_file_list_reset() to get it.
 *
 * @throws FileListException
 */
void
FileListRegistry::reset_file_list(const Glib::ustring& name,
                                  const Glib::RefPtr<Gtk::TextBuffer>& buf,
                                  file_list_factory_from_buffer factory)
{
  file_list_iter it = lists_.find(name);

  if (it != lists_.end())
  {
    SharedFileList list = factory(name, buf);

    // let task manager take care of the task which owns this list
    signal_file_list_reset_.emit(list);

    // now update all the list which contain it
    file_list_iter lists_it = lists_.begin();
    file_list_iter lists_end = lists_.end();

    for ( ; lists_it != lists_end; ++lists_it)
    {
      if (lists_it->second->contains_item(list->get_name()))
      {
        SharedItem cast_item = std::tr1::static_pointer_cast<Item>(list);
        lists_it->second->reset_item(cast_item);
      }
    }

    lists_[name] = list;
  }
  else
  {
    throw FileListNotFoundException(file_list_not_found_msg);
  }
}

/** Removes a list from the internal data structure and deletes the
 * associated file.
 *
 * @throw FileListNotFoundException
 */
void
FileListRegistry::delete_file_list(const Glib::ustring& name)
{
  file_list_iter it = lists_.find(name);

  if (it != lists_.end())
  {
    g_remove(it->second->get_path().c_str());

    lists_.erase(it);
  }
  else
  {
    throw FileListNotFoundException(file_list_not_found_msg);
  }
}

/**
 * @param name File list name.
 *
 * @return Path of the file where the list is defined.
 *
 * @throw FileListNotFoundException
 */
Glib::ustring
FileListRegistry::get_file_list_path(const Glib::ustring& name)
{
  file_list_iter it = lists_.find(name);

  if (it != lists_.end())
    return it->second->get_path();
  else
    throw FileListNotFoundException(file_list_not_found_msg);
}

/**
 * @return Whether a list 'name' is registered
 */
bool
FileListRegistry::list_name_in_use(const Glib::ustring& name) const
{
  file_list_const_iter it = lists_.find(name);

  if (it != lists_.end())
    return true;
  else
    return false;
}

/**
 * @return Signal emitted when a file list is reset, that is changes type
 * which requires a new object to be allocated.
 */
sigc::signal< void, SharedFileList& >&
FileListRegistry::get_signal_file_list_reset()
{
  return this->signal_file_list_reset_;
}

} // namespace TaskHandling

} // namespace USBSink
