/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 *  USBSink - filelistregistry.h
 *
 *  Copyright (C) 2006 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _TASK_FILE_LIST_REGISTRY_H_
#define _TASK_FILE_LIST_REGISTRY_H_

#include "exception.h"
#include "filelist.h"
#include "taskitem.h"
#include "uncopyable.h"

#include <list>
#include <map>
#include <memory>
#include <tr1/memory>

#include <glibmm/ustring.h>

#include <gtkmm/textbuffer.h>

namespace USBSink
{

namespace TaskHandling
{

typedef SharedFileList (*file_list_factory) (const Glib::ustring&);

typedef SharedFileList (*file_list_factory_from_buffer)
  (const Glib::ustring&,
   const Glib::RefPtr<Gtk::TextBuffer>&);


/** Indicates that a FileList, which the client has asked for, doesn't exist,
 * and should probably go and register a new one.
 */
class FileListNotFoundException : public Exception
{
 public:
  FileListNotFoundException(const Glib::ustring msg) : Exception(msg) {}
};


/** Indicates that a FileList, which the client has tried to register, already
 * exists, and should use the get_file_list() method to get it.
 */
class FileListExistsException : public Exception
{
 public:
  FileListExistsException(const Glib::ustring msg) : Exception(msg) {}
};


/** Interface for getting handles to existing and registering new file lists.
 * Whenever any part of code needs to perform this task, it should use
 * functions defined here.
 *
 * The intention is to prevent noise and confusion which could occur in
 * situations such as when one file list including another, and needs to
 * clean up, ie it is necessary to have one shared object per file list.
 *
 * This is a singleton, use the instance() method to access the object.
 */
class FileListRegistry : public Uncopyable
{
 public:

  static FileListRegistry* instance();
  ~FileListRegistry();

  static bool is_initialized();

  SharedItem  get_file_list(const Glib::ustring& name);

  void        get_file_lists(std::list<SharedFileList>& ret_list);

  SharedItem  register_new_file_list(const Glib::ustring& name,
                                     file_list_factory f);

  SharedItem  register_new_file_list(const Glib::ustring& name,
                                     const Glib::RefPtr<Gtk::TextBuffer>& buf,
                                     file_list_factory_from_buffer f);

  void        reset_file_list(const Glib::ustring& name,
                              const Glib::RefPtr<Gtk::TextBuffer>& buf,
                              file_list_factory_from_buffer f);

  void           delete_file_list(const Glib::ustring& name);

  Glib::ustring  get_file_list_path(const Glib::ustring& name);

  bool           list_name_in_use(const Glib::ustring& name) const;

  sigc::signal< void, SharedFileList& >& get_signal_file_list_reset();

 protected:

  FileListRegistry();

  void load_task_file_lists();

  sigc::signal< void, SharedFileList& > signal_file_list_reset_;

 private:

  static bool initialized_;
  static std::auto_ptr<FileListRegistry> instance_;

  std::map<Glib::ustring, SharedFileList> lists_;
  // TODO: add Glib::ustring::compare?
};

// Implementation of FileListRegistry's inline functions

inline bool
FileListRegistry::is_initialized()
{
  return initialized_;
}

} // namespace TaskHandling

} // namespace USBSink

#endif // _TASK_FILE_LIST_REGISTRY_H_
