/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 *  USBSink - halclient.cc
 *
 *  Copyright (C) 2005 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "common/nmv-log-stream-utils.h"
#include "flashdrive.h"
#include "halclient.h"

#include <string.h>

#include <algorithm>
#include <memory>
#include <string>
#include <vector>

#include <tr1/memory>

#include <glib/gmessages.h>

#include <dbus/dbus-glib-lowlevel.h>

#include <libhal.h>

#include <sigc++/sigc++.h>

using std::auto_ptr;
using std::string;
using std::vector;

using std::tr1::shared_ptr;

namespace
{

static bool
is_acceptable_bus(LibHalDriveBus bus)
{
  switch (bus)
  {
  case LIBHAL_DRIVE_BUS_USB:
  case LIBHAL_DRIVE_BUS_SCSI:
    return true;
  default:
    return false;
  }
}

static bool
is_acceptable_drive_type(LibHalDriveType drive_type)
{
  switch (drive_type)
  {
  case LIBHAL_DRIVE_TYPE_REMOVABLE_DISK:
  case LIBHAL_DRIVE_TYPE_COMPACT_FLASH:
  case LIBHAL_DRIVE_TYPE_MEMORY_STICK:
  case LIBHAL_DRIVE_TYPE_SMART_MEDIA:
  case LIBHAL_DRIVE_TYPE_SD_MMC:
  case LIBHAL_DRIVE_TYPE_CAMERA:
  case LIBHAL_DRIVE_TYPE_PORTABLE_AUDIO_PLAYER:
  case LIBHAL_DRIVE_TYPE_ZIP:
  case LIBHAL_DRIVE_TYPE_JAZ:
  case LIBHAL_DRIVE_TYPE_FLASHKEY:
    return true;
  default:
    return false;
  }
}

// Libhal callbacks.

// When this function is called, the volume is not mounted by a volume
// manager yet, but it needs to be available to the user.
// Likewise, on_removed executes before unmount.
// (Un)mounting triggers on_hal_property_modified.
static void
on_hal_device_added(LibHalContext* ctx, const char* udi)
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;
  
  LibHalVolume* volume;

  if ((volume = libhal_volume_from_udi(ctx, udi)))
  {
    LibHalDrive* drive =
      libhal_drive_from_udi(ctx, libhal_volume_get_storage_device_udi(volume));

    LibHalDriveType drive_type = libhal_drive_get_type(drive);
    LibHalDriveBus drive_bus = libhal_drive_get_bus(drive);

    LOG_DD("new device " << udi <<
           " of type " <<
           USBSink::HalClient::hal_drive_type_to_string(drive_type) <<
           " added on bus " <<
           USBSink::HalClient::hal_bus_to_string(drive_bus));
  
    if (is_acceptable_bus(drive_bus) && is_acceptable_drive_type(drive_type))
    {
      USBSink::HalClient* hc = USBSink::HalClient::instance();
      hc->add_drive(drive, volume, drive_bus);
    }
    libhal_drive_free(drive);
  }
  libhal_volume_free(volume);
}

// There are two ways for a volume to become unavailable:
// when the user unplugs a drive with a mounted volume and when she
// unmounts the volume, but the drive stays plugged in.

// This function is executed when a drive is unplugged from the computer,
// and it is then when we need to remove it from the drives_ vector.
// Unmounting a volume should stop all related tasks, but we need to keep
// watching it.
static void
on_hal_device_removed(LibHalContext* /* ctx */, const char* _udi)
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  LOG_DD("removed " << _udi);

  string udi(_udi);

  USBSink::HalClient* hc = USBSink::HalClient::instance();

  USBSink::SharedFlashDrive flash = hc->find_flash_by_udi(udi);

  if (flash)
  {
    LOG_DD("removing flash object " << flash->get_label());
    hc->remove_drive(flash);
  }
  else
  {
    LOG_ERROR("could not get a pointer to a FlashDrive for " << udi);
  }
}

// When a drive is plugged or unplugged, first changes the volume.mount_point
// and then volume.is_mounted property. If there is a volume manager, that is.
// For tasks and gui, what matters are the mount/unmount events, so appropriate
// signals are emitted here.
static void
on_hal_property_modified(LibHalContext* ctx,
			 const char* _udi,
			 const char* key,
			 dbus_bool_t /* is_removed */,
			 dbus_bool_t /* is_added */)
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  LOG_DD("something's changed on " << _udi << ": " << key);

  if (strcmp(key, "volume.is_mounted") == 0)
  {
    USBSink::HalClient* hc = USBSink::HalClient::instance();
    string udi(_udi);
    USBSink::SharedFlashDrive flash = hc->find_flash_by_udi(udi);

    if (flash)
    {
      LOG_DD("getting volume from udi");
      LibHalVolume* volume = libhal_volume_from_udi(ctx, _udi);

      if ((volume == NULL) || (! libhal_volume_is_mounted(volume)))
      {
        LOG_DD("device unmounted; emitting signal");
        flash->disable();
        hc->emit_signal_on_flash(flash, "unmounted");
      }
      else
      {
        LOG_DD("getting storage device + drive");
        LibHalDrive* drive =
          libhal_drive_from_udi(ctx,
                                libhal_volume_get_storage_device_udi(volume));

        LOG_DD("device mounted; emitting signal");
    
        if (libhal_volume_is_mounted(volume))
        {
          flash->load_dev_info(drive, volume);
          hc->emit_signal_on_flash(flash, "mounted");
        }
    
        libhal_drive_free(drive);
      }

      libhal_volume_free(volume);
    }
    else
    {
      LOG_ERROR("could not get a pointer to FlashDrive for " << _udi);
    }
  }
}

} // anonymous namespace

namespace USBSink
{

auto_ptr<HalClient> HalClient::instance_;

HalClient*
HalClient::instance()
{
  if (! instance_.get())
  {
    instance_.reset(new HalClient);
  }
  return instance_.get();
}

HalClient::HalClient()
{
  if (! init_hal())
  {
    LOG_ERROR("Can't connect to HAL! Please start hald.");
    throw HalError("Error initializing HAL.");
  }
}

HalClient::~HalClient()
{
  disconnect_from_hal();
  drives_.clear();
}

bool
HalClient::init_hal()
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  if (! (hal_ctx_ = libhal_ctx_new()))
  {
    LOG_ERROR("Failed to create new HAL context!");
    return false;
  }
  
  DBusError error;
  dbus_error_init(&error);

  if (! integrate_hal_into_mainloop(&error))
  {
    LOG_ERROR("Failed to integrate HAL into mainloop: " << error.message);
    dbus_error_free(&error);
    return false;
  }

  set_hal_callbacks();
  
  if (! libhal_device_property_watch_all(hal_ctx_, &error))
  {
    LOG_ERROR("Failed to watch all HAL properties: " << error.message);
    dbus_error_free(&error);
    return false;
  }
  
  if (! libhal_ctx_init(hal_ctx_, &error))
  {
    LOG_ERROR("Failed to initialize HAL: " << error.message);
    dbus_error_free(&error);
    return false;
  }

  if (! get_drives_at_startup(&error))
  {
    LOG_ERROR("Failed to get info about currently available devices: "
	      << error.message);
    dbus_error_free(&error);
    return false;
  }
	
  return true;
}

bool
HalClient::integrate_hal_into_mainloop(DBusError* error)
{
  dbus_conn_ = dbus_bus_get(DBUS_BUS_SYSTEM, error);
  
  if (dbus_error_is_set(error))
  {
    return false;
  }
  
  dbus_connection_setup_with_g_main(dbus_conn_, NULL);
  
  libhal_ctx_set_dbus_connection(hal_ctx_, dbus_conn_);
  
  return true;
}

void
HalClient::set_hal_callbacks()
{
  libhal_ctx_set_device_added(hal_ctx_, on_hal_device_added);
  libhal_ctx_set_device_removed (hal_ctx_, on_hal_device_removed);
  libhal_ctx_set_device_property_modified(hal_ctx_, on_hal_property_modified);
}

// Should be called only in the ctor.
bool
HalClient::get_drives_at_startup(DBusError* error)
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  char** devices;
  int num_devices;

  LOG_DD("looking for drives");
  
  devices = libhal_get_all_devices(hal_ctx_, &num_devices, error);
  if (dbus_error_is_set(error))
  {
    return false;
  }
  else
  {
    for (int i = 0; i < num_devices; ++i)
    {
      LibHalDrive* drive;
      if ((drive = libhal_drive_from_udi(hal_ctx_, devices[i])))
      {
	// The device is a drive.
        // Comparing drive type against LIBHAL_DRIVE_TYPE_REMOVABLE_DISK
        // is not good because, for instance, external HD is not a removable.

        LibHalDriveType drive_type = libhal_drive_get_type(drive);
        LibHalDriveBus drive_bus = libhal_drive_get_bus(drive);

        LOG_DD("found " << devices[i] <<
               ": a " <<
               USBSink::HalClient::hal_drive_type_to_string(drive_type) <<
               " drive on bus " <<
               USBSink::HalClient::hal_bus_to_string(drive_bus));

        if (is_acceptable_bus(drive_bus) &&
            is_acceptable_drive_type(drive_type))
	{
	  char** volumes;
	  int num_volumes;
	  
	  if ((volumes = libhal_drive_find_all_volumes(hal_ctx_,
						       drive,
						       &num_volumes))
	      && (num_volumes))
	  {
            for (int vol_index = 0; vol_index < num_volumes; ++vol_index)
            {
              LibHalVolume* volume =
                libhal_volume_from_udi(hal_ctx_, volumes[vol_index]);
	    
              SharedFlashDrive flash(
                new FlashDrive(dbus_conn_, drive, volume, drive_bus));
              drives_.push_back(flash);
	    
              libhal_volume_free(volume);
            }

            libhal_free_string_array(volumes);
	  }
          else
          {
            LOG_ERROR("Could not determine volumes on drive " << devices[i]);
          }
	}
      }
      libhal_drive_free(drive);
    }
  }
  libhal_free_string_array(devices);
  
  return true;
}

void
HalClient::disconnect_from_hal()
{
  libhal_ctx_shutdown(hal_ctx_, NULL);
  libhal_ctx_free(hal_ctx_);
}

void
HalClient::add_drive(LibHalDrive* drive,
                     LibHalVolume* volume,
                     LibHalDriveBus bus)
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  SharedFlashDrive flash(new FlashDrive(dbus_conn_, drive, volume, bus));
  drives_.push_back(flash);
  signal_added.emit(flash);
}

void
HalClient::remove_drive(SharedFlashDrive flash)
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  signal_removed.emit(flash);
  drives_.erase(remove(drives_.begin(), drives_.end(), flash),
		drives_.end());
}

SharedFlashDrive
HalClient::find_flash_by_udi(string& udi)
{
  FlashDrive* fd = 0;
  SharedFlashDrive sfd(fd);
  FlashDriveVector::iterator it = drives_.begin();
  
  for (; it != drives_.end(); ++it)
  {
    if ((*it)->get_udi() == udi)
    {
      sfd = *it;
      break;
    }
  }
  return sfd;
}

void
HalClient::emit_signal_on_flash(SharedFlashDrive sfd, const char* signal_name)
{
  if (strcmp(signal_name, "mounted") == 0)
  {
    signal_mounted.emit(sfd);
  }
  else if (strcmp(signal_name, "unmounted") == 0)
  {
    signal_unmounted.emit(sfd);
  }
}

char*
HalClient::hal_drive_type_to_string(LibHalDriveType drive_type)
{
  switch (drive_type)
  {
  case LIBHAL_DRIVE_TYPE_REMOVABLE_DISK:
    return "removable disk";
  case LIBHAL_DRIVE_TYPE_DISK:
    return "disk";
  case LIBHAL_DRIVE_TYPE_CDROM:
    return "CDROM";
  case LIBHAL_DRIVE_TYPE_FLOPPY:
    return "floppy";
  case LIBHAL_DRIVE_TYPE_TAPE:
    return "tape";
  case LIBHAL_DRIVE_TYPE_COMPACT_FLASH:
    return "compact flash";
  case LIBHAL_DRIVE_TYPE_MEMORY_STICK:
    return "memory stick";
  case LIBHAL_DRIVE_TYPE_SMART_MEDIA:
    return "smart media";
  case LIBHAL_DRIVE_TYPE_SD_MMC:
    return "SD/MMC";
  case LIBHAL_DRIVE_TYPE_CAMERA:
    return "camera";
  case LIBHAL_DRIVE_TYPE_PORTABLE_AUDIO_PLAYER:
    return "portable audio player";
  case LIBHAL_DRIVE_TYPE_ZIP:
    return "ZIP";
  case LIBHAL_DRIVE_TYPE_JAZ:
    return "JAZ";
  case LIBHAL_DRIVE_TYPE_FLASHKEY:
    return "flash key";
  case LIBHAL_DRIVE_TYPE_MO:
    return "MO";
  default:
    return "unknown";
  }
}

char*
HalClient::hal_bus_to_string(LibHalDriveBus bus_type)
{
  switch (bus_type)
  {
  case LIBHAL_DRIVE_BUS_IDE:
    return "IDE";
  case LIBHAL_DRIVE_BUS_SCSI:
    return "SCSI";
  case LIBHAL_DRIVE_BUS_USB:
    return "USB";
  case LIBHAL_DRIVE_BUS_IEEE1394:
    return "IEEE 1394 (FireWire)";
  case LIBHAL_DRIVE_BUS_CCW:
    return "CCW";
  default:
    return "unknown";
  }
}

} // namespace USBSink
