/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 *  USBSink - mainwindow.cc
 *
 *  Copyright (C) 2006-2007 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "common/nmv-log-stream-utils.h"
#include "aboutdialog.h"
#include "edittaskdialog.h"
#include "filelisteditor.h"
#include "flashdrivesdialog.h"
#include "globalstrings.h"
#include "iconloader.h"
#include "logviewdialog.h"
#include "mainwindow.h"
#include "newtaskdialog.h"
#include "taskmanager.h"
#include "utils.h"

#include <config.h>

#include <list>

#include <glib/gi18n.h>

#include <gtk/gtkmessagedialog.h>
#include <gtk/gtkstatusicon.h>
#include <gtk/gtkwindow.h>

#include <libgnome/gnome-help.h>

using std::list;
using USBSink::TaskHandling::SharedTask;

typedef USBSink::TaskHandling::Manager TaskManager;


namespace USBSink
{

namespace UI
{

static const gint DEFAULT_WINDOW_WIDTH = 540;
static const gint DEFAULT_WINDOW_HEIGHT = 300;


MainWindow::MainWindow(bool minimized)
:
  vbox_top_          (0),
  vbox_menu_toolbar_ (0),
  main_menu_         (0),
  toolbar_           (0),
  task_view_         (0),
  statusbar_         (0),
  hide_in_tray_      (minimized),
  last_window_pos_   ()
{
  init_ui();

  connect_signals();

  //TaskManager::init();
}

MainWindow::~MainWindow()
{
}

void
MainWindow::load_xml()
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  using Gnome::Glade::Xml;

  try
  {
    const Glib::RefPtr<Xml> xml = Xml::create(glade_mainwindow_filename);
    
    Gtk::Window* dummy_window = 0;
    window_.reset(xml->get_widget("mainwindow", dummy_window));

    xml->get_widget("vbox_top", vbox_top_);
    xml->get_widget("vbox_menu_toolbar", vbox_menu_toolbar_);
    xml->get_widget("task_view", task_view_);
    xml->get_widget("statusbar", statusbar_);
  }
  catch (const Gnome::Glade::XmlError& xml_err)
  {
    LOG_EXCEPTION_DD(xml_err.what());
  }
}

void
MainWindow::connect_signals()
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  using sigc::mem_fun;

  window_->signal_hide().connect(
    mem_fun(*this, &MainWindow::on_hide));

  window_->signal_hide().connect(
    sigc::bind(
      mem_fun(*(TaskManager::instance()), &TaskManager::on_application_exit),
      this));

  window_->signal_show().connect(
    mem_fun(*this, &MainWindow::on_show));

  window_->signal_key_release_event().connect(
    mem_fun(*this, &MainWindow::on_key_release_event));

  // task manager
  TaskManager* manager = TaskManager::instance();

  manager->get_signal_task_added().connect(
    mem_fun(*this, &MainWindow::on_task_added));

  manager->get_signal_task_removed().connect(
    mem_fun(*this, &MainWindow::on_task_removed));

  manager->load_tasks();

  // connect to each task's signals
  list<SharedTask> tasks = manager->get_tasks();

  list<SharedTask>::iterator it(tasks.begin());
  list<SharedTask>::iterator end(tasks.end());

  for ( ; it != end; ++it)
  {
    (*it)->signal_begin_sync.connect(
      mem_fun(*this, &MainWindow::on_task_begin_sync));

    (*it)->signal_end_sync.connect(
      mem_fun(*this, &MainWindow::on_task_end_sync));
  }

  // halclient signals

  try
  {
    HalClient* hc = HalClient::instance();
    
    hc->signal_mounted.connect(
      mem_fun(*this, &MainWindow::on_drive_mounted));
    
    hc->signal_unmounted.connect(
      mem_fun(*this, &MainWindow::on_drive_unmounted));
  }
  catch (HalError& err)
  {
    LOG_EXCEPTION_DD("Hal exception: " << err.what());
  }
}

/**
 * Initializes action group and all actions.
 */
void
MainWindow::init_actions()
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  using sigc::mem_fun;

  action_group_ = Gtk::ActionGroup::create();
  LOG_DD("Created ActionGroup");

  // File menu
  action_group_->add(Gtk::Action::create("MenuFile", _("_File")));

  action_group_->add(
    Gtk::Action::create("NewTask", Gtk::Stock::NEW, _("_New Task")),
    mem_fun(*this, &MainWindow::on_new_task));

  action_group_->add(
    Gtk::Action::create("Quit", Gtk::Stock::QUIT),
    mem_fun(*this, &MainWindow::on_quit));

  // Edit menu
  action_group_->add(Gtk::Action::create("MenuEdit", _("_Edit")));

  action_group_->add(
    Gtk::Action::create("EditTask", Gtk::Stock::EDIT, _("_Task Properties")),
    mem_fun(*this, &MainWindow::on_edit_task));

  action_group_->add(
    Gtk::Action::create("DeleteTask", Gtk::Stock::DELETE),
    mem_fun(*this, &MainWindow::on_delete_task));

  action_group_->add(
    Gtk::Action::create("SelectAllTasks", _("Select _All")),
    mem_fun(*this, &MainWindow::on_select_all));

  // View menu
  action_group_->add(Gtk::Action::create("MenuView", _("_View")));

  action_group_->add(
    Gtk::RadioAction::create(view_tasks_radio_group_,
                             "ViewAllTasks",
                             _("_All Tasks")),
    mem_fun(*this, &MainWindow::on_show_all_tasks));

  action_group_->add(
    Gtk::RadioAction::create(view_tasks_radio_group_,
                             "ViewRunningTasks",
                             _("_Running Tasks")),
    mem_fun(*this, &MainWindow::on_show_running_tasks));

  action_group_->add(
    Gtk::RadioAction::create(view_tasks_radio_group_,
                             "ViewStoppedTasks",
                             _("_Stopped Tasks")),
    mem_fun(*this, &MainWindow::on_show_stopped_tasks));

  action_group_->add(
    Gtk::Action::create("ViewFlashDrives", _("Available _Flash Drives")),
    mem_fun(*this, &MainWindow::on_available_drives));

  action_group_->add(
    Gtk::Action::create("ViewLists", _("Lists")),
    mem_fun(*this, &MainWindow::on_view_file_list_editor));

  action_group_->add(
    Gtk::Action::create("ViewTransferLog", _("Transfer _Log")),
    Gtk::AccelKey("<control>L"),
    mem_fun(*this, &MainWindow::on_view_log));

  // Actions menu
  action_group_->add(Gtk::Action::create("MenuActions", _("_Actions")));

  action_group_->add(
    Gtk::Action::create("StartTasks", Gtk::Stock::YES, _("_Start")),
    Gtk::AccelKey("<control>Return"),
    mem_fun(*this, &MainWindow::on_start));

  action_group_->add(
    Gtk::Action::create("StopTasks", Gtk::Stock::STOP, _("S_top")),
    Gtk::AccelKey("Escape"),
    mem_fun(*this, &MainWindow::on_stop));

  // Help menu
  action_group_->add(Gtk::Action::create("MenuHelp", _("_Help")));

  action_group_->add(
    Gtk::Action::create("HelpContents", Gtk::Stock::HELP),
    mem_fun(*this, &MainWindow::on_help));

  action_group_->add(
    Gtk::Action::create("AboutProgram", Gtk::Stock::ABOUT),
    mem_fun(*this, &MainWindow::on_about));

  LOG_DD("Created actions");
}

/**
 * Initializes UIManager with the action group and UI description file.
 */
void
MainWindow::init_ui_manager()
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  ui_manager_ = Gtk::UIManager::create();
  ui_manager_->insert_action_group(action_group_);
  window_->add_accel_group(ui_manager_->get_accel_group());
  
  try
  {
    ui_manager_->add_ui_from_file(USBSink::ui_description_path);
    LOG_DD("Added UI description");
  }
  catch (const Glib::Error& ex)
  {
    // We can't do anything without the UI definition.
    LOG_EXCEPTION_DD(ex.what());
    throw ex;
  }
}

/**
 * Loads menu and toolbar objects from UIManager, packs them in the VBoxes.
 */
void
MainWindow::show_menu_toolbar()
{
  // Get the menu and toolbar objects from UIManager.
  main_menu_ =
    static_cast<Gtk::Menu*>(ui_manager_->get_widget("/MainMenu"));
  g_assert(main_menu_);

  toolbar_ =
    static_cast<Gtk::Toolbar*>(ui_manager_->get_widget("/MainToolbar"));
  g_assert(toolbar_);

  // Pack them on their VBox.
  vbox_menu_toolbar_->pack_start(*main_menu_, Gtk::PACK_SHRINK);

  toolbar_->set_toolbar_style(Gtk::TOOLBAR_BOTH);
  vbox_menu_toolbar_->pack_start(*toolbar_, Gtk::PACK_SHRINK, 0);
}

/** \brief Main user interface loading function.
 *
 * Calls all the small functions in correct order.
 */
void
MainWindow::init_ui()
{
  load_xml();

  init_actions();

  init_ui_manager();

  show_menu_toolbar();

  setup_treeview();

  gtk_window_set_default_icon_name("usbsink");
  
  window_->set_default_size(DEFAULT_WINDOW_WIDTH, DEFAULT_WINDOW_HEIGHT);
}

void
MainWindow::setup_treeview()
{
  // create the tree model
  task_treemodel_ = Gtk::ListStore::create(task_columns_);
  task_view_->set_model(task_treemodel_);

  // set up the treeview selection object
  task_selection_ = task_view_->get_selection();
  task_selection_->set_mode(Gtk::SELECTION_MULTIPLE);

  // append column headers

  // for Status, display a CellRendererPixbuf instead of a CellRendererText
  Gtk::CellRendererPixbuf cell;
  task_view_->append_column(_("Status"), cell);

  Gtk::TreeViewColumn* column = task_view_->get_column(0); // the status column
  column->add_attribute(cell.property_stock_id(), task_columns_.col_status);

  task_view_->append_column(_("Task name"),
                            task_columns_.col_name);

  task_view_->append_column(_("Removable drive"),
                            task_columns_.col_drive_label);

  task_view_->append_column(_("Transferred"),
                            task_columns_.col_bytes_transferred);
}

void
MainWindow::on_new_task()
{
  if (new_task_dialog_.get())
  {
     // The dialog is already open, and the user asked for it once again;
     // so move it where the user can see it.
     // Btw present() calls Widget::show() if a window is hidden.
     new_task_dialog_->get_dialog()->present();
  }
  else
  {
    std::auto_ptr<NewTaskDialog> dialog(new NewTaskDialog(*window_));
    
    dialog->get_dialog()->signal_hide().connect(
      sigc::mem_fun(*this, &MainWindow::on_new_task_hide));

    dialog->signal_file_list_editor_request.connect(
      sigc::mem_fun(*this, &MainWindow::on_file_list_editor));

    dialog->get_dialog()->show();
    
    new_task_dialog_ = dialog;
  }
}

void
MainWindow::on_new_task_hide()
{
  // Transfer ownership and let the destructor do the delete.
  const std::auto_ptr<NewTaskDialog> temp(new_task_dialog_);
}

/**
 * Handler for a request to open the file list editor.
 */
void
MainWindow::on_view_file_list_editor()
{
  if (file_list_editor_.get())
  {
     file_list_editor_->get_dialog()->present();
  }
  else
  {
    std::auto_ptr<FileListEditor> dialog(new FileListEditor(*window_));
    
    dialog->get_dialog()->signal_hide().connect(
      sigc::mem_fun(*this, &MainWindow::on_file_list_editor_hide));
    dialog->get_dialog()->show();
    
    file_list_editor_ = dialog;
  }
}

/**
 * Handler for a request to open the file list editor and to register a new
 * list under the specified name.
 */
void
MainWindow::on_file_list_editor(const Glib::ustring& list_name)
{
  if (file_list_editor_.get())
  {
     file_list_editor_->get_dialog()->present();
  }
  else
  {
    std::auto_ptr<FileListEditor> dialog;

    if (list_name != Glib::ustring())
    {
      dialog.reset(new FileListEditor(*window_, list_name));
    }
    else
    {
      dialog.reset(new FileListEditor(*window_));
    }
    
    dialog->get_dialog()->signal_hide().connect(
      sigc::mem_fun(*this, &MainWindow::on_file_list_editor_hide));
    dialog->get_dialog()->show();
    
    file_list_editor_ = dialog;
  }
}

void
MainWindow::on_file_list_editor_hide()
{
  const std::auto_ptr<FileListEditor> temp(file_list_editor_);
}

void
MainWindow::on_edit_task()
{
  list<Gtk::TreeModel::Path> rows(task_selection_->get_selected_rows());

  if (! rows.empty())
  {
    Gtk::TreeModel::Row row = *(task_treemodel_->get_iter(*rows.begin()));

    TaskManager* tm = TaskManager::instance();
    SharedTask sel_task = tm->get_task(row[task_columns_.col_id]);

    if (edit_task_dialog_.get())
    {
      edit_task_dialog_->get_dialog()->present();
    }
    else
    {
      std::auto_ptr<EditTaskDialog> dialog(
	new EditTaskDialog(*window_, sel_task));
      
      dialog->get_dialog()->signal_hide().connect(
	sigc::mem_fun(*this, &MainWindow::on_edit_task_hide));

      dialog->signal_file_list_editor_request.connect(
        sigc::mem_fun(*this, &MainWindow::on_file_list_editor));
      
      dialog->get_dialog()->show();

      edit_task_dialog_ = dialog;
    }
  }
}

void
MainWindow::on_edit_task_hide()
{
  const std::auto_ptr<EditTaskDialog> temp(edit_task_dialog_);
}


void
MainWindow::on_delete_task()
{
  list<Gtk::TreeModel::Path> rows(task_selection_->get_selected_rows());

  if (! rows.empty())
  {
    Gtk::TreeModel::Row row = *(task_treemodel_->get_iter(*rows.begin()));

    int task_id = row[task_columns_.col_id];

    TaskManager* manager = TaskManager::instance();
    const Glib::ustring task_name = manager->get_task_name(task_id);

    Gtk::MessageDialog dialog(
      *(get_window()),
      Util::i18n_printf(_("Are you sure you want to permanently delete '%s'?"),
                        task_name.c_str()),
      false /* use_markup */,
      Gtk::MESSAGE_WARNING,
      Gtk::BUTTONS_NONE);

    dialog.set_title(_("Delete task?"));

    Gtk::StockID stock_cancel(Gtk::Stock::CANCEL);
    Gtk::StockID stock_delete(Gtk::Stock::DELETE);

    dialog.add_button(stock_cancel, Gtk::RESPONSE_CANCEL);
    dialog.add_button(stock_delete, Gtk::RESPONSE_OK);

    dialog.set_default_response(Gtk::RESPONSE_OK);

    dialog.set_secondary_text(_("If you delete a task, its file and "
                                "associated list will be permanently lost."));

    int result = dialog.run();

    if (result == Gtk::RESPONSE_OK)
      manager->remove_task(task_id);
  }
}

void
MainWindow::on_select_all()
{
  task_selection_->select_all();
}

void
MainWindow::on_start()
{
  list<Gtk::TreeModel::Path> rows(task_selection_->get_selected_rows());

  if (! rows.empty())
  {
    list<Gtk::TreeModel::Path>::iterator it(rows.begin());
    list<Gtk::TreeModel::Path>::iterator end(rows.end());

    for ( ; it != end; ++it)
    {
      Gtk::TreeModel::Row row = *(task_treemodel_->get_iter(*it));

      TaskManager* tm = TaskManager::instance();
      SharedTask sel_task = tm->get_task(row[task_columns_.col_id]);

      if (sel_task->get_drive().get() == 0)
      {
	Glib::ustring message(
	  Util::i18n_printf(_("Cannot start task %s"),
	                    sel_task->get_name().c_str()));

	Gtk::MessageDialog err_dialog(*(get_window()),
				      message, false, Gtk::MESSAGE_ERROR,
				      Gtk::BUTTONS_OK, true);
	Glib::ustring err_details(_("Target drive is not present."));
	err_dialog.set_secondary_text(err_details);
	err_dialog.run();
      }
      else
      {
	if (! sel_task->get_drive()->is_mounted())
	{
	  Glib::ustring message(
	    Util::i18n_printf(_("Cannot start task %s"),
	                    sel_task->get_name().c_str()));
	  Gtk::MessageDialog err_dialog(*(get_window()),
					message, false, Gtk::MESSAGE_ERROR,
					Gtk::BUTTONS_OK, true);
	  
	  // TODO: suggest and offer actions with some buttons!!
	  Glib::ustring err_details(_("Target drive is not mounted."));
	  err_dialog.set_secondary_text(err_details);
	  err_dialog.run();
	}
	else
	{
	  if (! sel_task->is_running())
	  {
	    sel_task->start();
	  }
	}
      }
    }
  }
}

void
MainWindow::on_stop()
{
  list<Gtk::TreeModel::Path> rows(task_selection_->get_selected_rows());

  if (! rows.empty())
  {
    list<Gtk::TreeModel::Path>::iterator it(rows.begin());
    list<Gtk::TreeModel::Path>::iterator end(rows.end());

    for ( ; it != end; ++it)
    {
      Gtk::TreeModel::Row row = *(task_treemodel_->get_iter(*it));
      
      TaskManager* tm = TaskManager::instance();
      SharedTask sel_task = tm->get_task(row[task_columns_.col_id]);
      
      if (sel_task->is_running())
	sel_task->stop("");
    }
  }
}

void
MainWindow::on_view_log()
{
  if (log_view_dialog_.get())
  {
    log_view_dialog_->get_dialog()->present();
  }
  else
  {
    std::auto_ptr<LogViewDialog> dialog(new LogViewDialog(*window_));
    
    dialog->get_dialog()->signal_hide().connect(
      sigc::mem_fun(*this, &MainWindow::on_log_view_dialog_hide));
    dialog->get_dialog()->show();
    
    log_view_dialog_ = dialog;
  }
}

void
MainWindow::on_log_view_dialog_hide()
{
  const std::auto_ptr<LogViewDialog> temp(log_view_dialog_);
}

void
MainWindow::on_help()
{
  GError* error = NULL;

  gnome_help_display ("usbsink", NULL, &error);

  if (error)
  {
    GtkWidget *dialog;

    dialog = gtk_message_dialog_new (GTK_WINDOW (get_window()->gobj()),
                                     GTK_DIALOG_DESTROY_WITH_PARENT,
                                     GTK_MESSAGE_ERROR,
                                     GTK_BUTTONS_CLOSE,
                                     _("Could not display help for USBSink\n"
                                       "%s"),
                                     error->message);
    g_signal_connect_swapped (dialog, "response",
                              G_CALLBACK (gtk_widget_destroy),
                              dialog);
    gtk_widget_show (dialog);
    g_error_free (error);
  }
}

void
MainWindow::on_about()
{
  if (about_dialog_.get())
  {
    about_dialog_->present();
  }
  else
  {
    std::auto_ptr<Gtk::Dialog> dialog(AboutDialog::create(*window_));
    
    dialog->signal_hide().connect(
      sigc::mem_fun(*this, &MainWindow::on_about_dialog_hide));
    dialog->show();
    
    about_dialog_ = dialog;
  }
}

void
MainWindow::on_about_dialog_hide()
{
  const std::auto_ptr<Gtk::Dialog> temp(about_dialog_);
}

void
MainWindow::show_tasks(ShowTasks choice)
{
  task_treemodel_->clear();

  TaskManager* manager = TaskManager::instance();

  list<SharedTask> tasks = manager->get_tasks();

  list<SharedTask>::iterator it(tasks.begin());
  list<SharedTask>::iterator end(tasks.end());

  for ( ; it != end; ++it)
  {
    switch (choice)
    {
    case SHOW_TASKS_RUNNING:
      if ((*it)->is_running())
	display_task(*it);
      break;

    case SHOW_TASKS_STOPPED:
      if (! (*it)->is_running())
	display_task(*it);
      break;

    case SHOW_TASKS_ALL:
      display_task(*it);
      break;
    }
  }
}

void
MainWindow::on_show_all_tasks()
{
    show_tasks(SHOW_TASKS_ALL);
}

void
MainWindow::on_show_running_tasks()
{
    show_tasks(SHOW_TASKS_RUNNING);
}

void
MainWindow::on_show_stopped_tasks()
{
    show_tasks(SHOW_TASKS_STOPPED);
}

void
MainWindow::on_available_drives()
{
  if (flash_drives_dialog_.get())
  {
    flash_drives_dialog_->get_dialog()->present();
  }
  else
  {
    std::auto_ptr<FlashDrivesDialog> dialog(new FlashDrivesDialog(*window_));
    
    dialog->get_dialog()->signal_hide().connect(
      sigc::mem_fun(*this, &MainWindow::on_available_drives_hide));
    dialog->get_dialog()->show();
    
    flash_drives_dialog_ = dialog;
  }
}

void
MainWindow::on_available_drives_hide()
{
  const std::auto_ptr<FlashDrivesDialog> temp(flash_drives_dialog_);
}

void
MainWindow::on_quit()
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  LOG_DD("Application quitting on user request");

  hide_in_tray_ = false;
  window_->hide(); // in on_hide() we'll quit()
}

/**
 * Caller function should always check if returned iterator is valid.
 */
tree_children::iterator
MainWindow::get_tree_iter_for_task(int task_id)
{
  tree_children children = task_treemodel_->children();
  tree_children::iterator iter = children.begin();
  tree_children::iterator end = children.end();
  
  for (; iter != end; ++iter)
  {
    if ((*iter)[task_columns_.col_id] == task_id)
    {
      break;
    }
  }

  return iter;
}

void
MainWindow::display_task(const SharedTask& task)
{
  // add a new row to the treeview
  Gtk::TreeModel::Row row = *(task_treemodel_->append());

  row[task_columns_.col_id] = task->get_id();

  if ((task->get_start_time_option() == TaskHandling::START_TIME_OPTION_NOW) &&
      task->is_running())
  {
    // col_status is a stockID string
    row[task_columns_.col_status] = Gtk::Stock::YES.id;
  }
  else
  {
    row[task_columns_.col_status] = Gtk::Stock::STOP.id;
  }

  row[task_columns_.col_name] = task->get_name();

  if (task->get_drive() != 0)
  {
    if (task->get_drive()->is_mounted())
    {
      row[task_columns_.col_drive_label] = task->get_drive()->get_label();
    }
    else
    {
      // the drive is present, but unmounted
      row[task_columns_.col_drive_label] =
	Util::i18n_printf(_("%s: not mounted"),
	                  task->get_drive()->get_label().c_str());
    }
  }
  else
  {
    // the drive is not present
    row[task_columns_.col_drive_label] = _("Not present");
  }
}

void
MainWindow::on_task_added(const SharedTask& task)
{
  display_task(task);

  // connect to its signals
  task->signal_started.connect(
    sigc::mem_fun(*this, &MainWindow::on_task_started));

  task->signal_stopped.connect(
    sigc::mem_fun(*this, &MainWindow::on_task_stopped));
}

void
MainWindow::on_task_removed(int task_id)
{
  tree_children::iterator iter(get_tree_iter_for_task(task_id));

  if (iter)
  {
    // erase the corresponding row
    task_treemodel_->erase(iter);
  }
}

void
MainWindow::on_task_started(TaskHandling::Task* task)
{
  tree_children::iterator iter(get_tree_iter_for_task(task->get_id()));

  if (iter)
  {
    Gtk::TreeModel::Row row = *iter;
 
    row[task_columns_.col_status] = Gtk::Stock::YES.id;
    row[task_columns_.col_drive_label] = task->get_drive()->get_label();
  }
}

void
MainWindow::on_task_stopped(TaskHandling::Task* task,
                            const Glib::ustring& note)
{
  tree_children::iterator iter(get_tree_iter_for_task(task->get_id()));

  if (iter)
  {
    Gtk::TreeModel::Row row = *iter;
 
    row[task_columns_.col_status] = Gtk::Stock::STOP.id;

    Glib::ustring label(row[task_columns_.col_drive_label]);
    if (note != Glib::ustring())
      label += " : " + note;

    row[task_columns_.col_drive_label] = label;
  }
}

void
MainWindow::on_task_begin_sync(TaskHandling::Task* task)
{
  statusbar_->push(Util::i18n_printf(_("Synchronizing files for %s"),
                                     task->get_name().c_str()));

  tree_children::iterator iter(get_tree_iter_for_task(task->get_id()));

  if (iter)
  {
    Gtk::TreeModel::Row row = *iter;
    row[task_columns_.col_status] = Gtk::Stock::EXECUTE.id;
  }
}

void
MainWindow::on_task_end_sync(TaskHandling::Task* task)
{
  statusbar_->push(Util::i18n_printf(_("Synchronized files for %s"),
                                     task->get_name().c_str()));

  tree_children::iterator iter(get_tree_iter_for_task(task->get_id()));

  if (iter)
  {
    Gtk::TreeModel::Row row = *iter;

    row[task_columns_.col_status] = Gtk::Stock::YES.id;
    row[task_columns_.col_bytes_transferred] =
      task->get_printed_bytes_transferred();
  }
}

void
MainWindow::on_drive_mounted(const SharedFlashDrive& flash)
{
  statusbar_->push(Util::i18n_printf(_("Drive %s mounted"),
                                     flash->get_label().c_str()));
}

void
MainWindow::on_drive_unmounted(const SharedFlashDrive& flash)
{
  statusbar_->push(Util::i18n_printf(_("Drive %s unmounted"),
                                     flash->get_label().c_str()));
}

static void
status_icon_activated_cb (GtkStatusIcon *icon, gpointer data)
{
  MainWindow* main_window = static_cast<MainWindow*>(data);

  gtk_status_icon_set_visible (icon, FALSE);

  main_window->get_window()->show_all();
}

void
MainWindow::on_hide()
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  if (hide_in_tray_)
  {
    LOG_DD("Minimizing to system tray");

    TaskManager* manager = TaskManager::instance();
    int active_tasks = manager->get_active_task_count();

    // Loading a large icon assures that it will be big enough on large panels
    // (it will be automatically resized to fit smaller ones anyway).
    // Currently vice-versa does not happen on gnome-panel.
    status_icon_ =
      Gtk::StatusIcon::create(IconLoader::load_program_icon(ICON_SIZE_LARGE));

    if (! active_tasks)
      status_icon_->set_tooltip(Util::i18n_printf(_("No active tasks")));
    else
      status_icon_->set_tooltip(
        Util::i18n_printf(
          (active_tasks > 1) ? _("%d active tasks") : _("%d active task"),
          active_tasks));

    GtkStatusIcon *c_icon = status_icon_->gobj();

    g_signal_connect (c_icon,
                      "activate",
                      G_CALLBACK (status_icon_activated_cb),
                      this);
  }
  else
  {
    LOG_DD("calling gtk::main::quit()");
    Gtk::Main::quit();
  }
}

void
MainWindow::on_show()
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  LOG_DD("Main window shown");

  hide_in_tray_ = true;

  LOG_DD("Moving window position to " << 
         last_window_pos_.x << ", " << last_window_pos_.y);
  window_->move(last_window_pos_.x, last_window_pos_.y);
}

bool
MainWindow::on_key_release_event(GdkEventKey* /* event */)
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  window_->get_position(last_window_pos_.x, last_window_pos_.y);
  LOG_DD("Saved window position to " << 
         last_window_pos_.x << ", " << last_window_pos_.y);

  return false; // propagate event further
}

} // namespace UI

} // namespace USBSink
