/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 *  USBSink - mainwindow.h
 *
 *  Copyright (C) 2006-2007 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _MAINWINDOW_H_
#define _MAINWINDOW_H_

#include "flashdrive.h"
#include "task.h"
#include "utils.h"

#include <memory>

#include <tr1/memory>

#include <gtkmm.h>
#include <sigc++/sigc++.h>
#include <libglademm/xml.h>

namespace USBSink
{

namespace UI
{
  class EditTaskDialog;
  class FlashDrivesDialog;
  class FileListEditor;
  class LogViewDialog;
  class NewTaskDialog;
}

namespace UI
{

enum ShowTasks
{
  SHOW_TASKS_ALL,
  SHOW_TASKS_RUNNING,
  SHOW_TASKS_STOPPED
};

class MainWindow : public sigc::trackable
{
public:
  
  explicit MainWindow(bool minimized = false);
  virtual ~MainWindow();

  Gtk::Window* get_window() { return window_.get(); }

  bool get_hide_in_tray() const;
  void set_hide_in_tray(bool hide_in_tray);
  
protected:

  class ModelColumns : public Gtk::TreeModel::ColumnRecord
  {
  public:
    ModelColumns()
      { add(col_id);
        add(col_status);
        add(col_name);
        add(col_drive_label);
        add(col_bytes_transferred);
      }

    Gtk::TreeModelColumn<int> col_id;
    Gtk::TreeModelColumn<Glib::ustring> col_status; // stock icon id
    Gtk::TreeModelColumn<Glib::ustring> col_name;
    Gtk::TreeModelColumn<Glib::ustring> col_drive_label;
    Gtk::TreeModelColumn<Glib::ustring> col_bytes_transferred;
  };

  // member functions
  void load_xml();
  void connect_signals();

  void init_actions();
  void init_ui_manager();
  void show_menu_toolbar();

  void setup_treeview();
  
  void init_ui();
  
  Gtk::TreeModel::Children::iterator get_tree_iter_for_task(int id);

  void display_task(const USBSink::TaskHandling::SharedTask& task);

  void show_tasks(ShowTasks choice);

  // signal handlers

  void on_hide();
  void on_show();

  bool on_key_release_event(GdkEventKey* event);

  // ui
  void on_new_task();
  void on_new_task_hide();

  void on_view_file_list_editor();
  void on_file_list_editor(const Glib::ustring& list_name);
  void on_file_list_editor_hide();
  
  void on_edit_task();
  void on_edit_task_hide();
  void on_delete_task();

  void on_select_all();
  
  void on_start();
  void on_stop();

  void on_view_log();
  void on_log_view_dialog_hide();

  void on_show_all_tasks();
  void on_show_running_tasks();
  void on_show_stopped_tasks();
  void on_available_drives();
  void on_available_drives_hide();

  void on_help();  
  void on_about();
  void on_about_dialog_hide();

  void on_quit();

  // application
  void on_task_added(const USBSink::TaskHandling::SharedTask& task);

  void on_task_removed(int task_id);

  void on_task_started(USBSink::TaskHandling::Task* t);

  void on_task_stopped(USBSink::TaskHandling::Task* t,
                       const Glib::ustring& note);

  void on_task_begin_sync(USBSink::TaskHandling::Task* t);

  void on_task_end_sync(USBSink::TaskHandling::Task* t);

  void on_drive_mounted(const SharedFlashDrive& sfd);

  void on_drive_unmounted(const SharedFlashDrive& sfd);

  // window pointer
  std::auto_ptr<Gtk::Window> window_;

  // dialogs
  std::auto_ptr<NewTaskDialog>     new_task_dialog_;
  std::auto_ptr<EditTaskDialog>    edit_task_dialog_;
  std::auto_ptr<Gtk::Dialog>       about_dialog_;
  std::auto_ptr<FlashDrivesDialog> flash_drives_dialog_;
  std::auto_ptr<FileListEditor>    file_list_editor_;
  std::auto_ptr<LogViewDialog>     log_view_dialog_;

  // top container
  Gtk::VBox* vbox_top_;

  // menu and toolbar items
  Glib::RefPtr<Gtk::UIManager> ui_manager_;
  Glib::RefPtr<Gtk::ActionGroup> action_group_;

  Gtk::VBox* vbox_menu_toolbar_;

  Gtk::RadioButtonGroup view_tasks_radio_group_;
  Gtk::Menu* main_menu_;
  Gtk::Toolbar* toolbar_;

  // treeview
  ModelColumns task_columns_;
  Glib::RefPtr<Gtk::ListStore> task_treemodel_;
  Gtk::TreeView* task_view_;
  Glib::RefPtr<Gtk::TreeSelection> task_selection_;

  // status bar
  Gtk::Statusbar* statusbar_;

  bool hide_in_tray_;
  Glib::RefPtr<Gtk::StatusIcon> status_icon_;

  USBSink::Util::Point last_window_pos_;
};


// inliners

inline bool
MainWindow::get_hide_in_tray() const
{
  return hide_in_tray_;
}

inline void
MainWindow::set_hide_in_tray(bool hide_in_tray)
{
  hide_in_tray_ = hide_in_tray;
}

} // namespace UI

} // namespace USBSink

#endif
