/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 *  USBSink - task.cc
 *
 *  Copyright (C) 2006-2007 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "common/nmv-log-stream-utils.h"
#include "dirtransfer.h"
#include "fileentry.h"
#include "filetransfer.h"
#include "halclient.h"
#include "task.h"
#include "transferstats.h"
#include "utils.h"

#include <iostream>
#include <string>

#include <glib/gi18n.h>

#include <glibmm/iochannel.h>
#include <glibmm/ustring.h>

#include <pcrecpp.h>

using std::list;

namespace USBSink
{

namespace TaskHandling
{

typedef list< Glib::RefPtr<Gnome::Vfs::Uri> > uri_list;
typedef uri_list::iterator uri_list_iter;
typedef uri_list::const_iterator uri_list_const_iter;

Task::Task(const int id,
           const Glib::ustring& task_name,
           const std::string& drive_udi,
           const DataFlow data_flow,
           const SharedItem& file_list,
           const StartTimeOption when_to_start,
           const int timeout,
           const uri_list& src_files,
           const Glib::ustring& target_dir)
  :
  id_(id),
  name_(task_name),
  drive_udi_(drive_udi),
  data_flow_(data_flow),
  file_list_(std::tr1::static_pointer_cast<FileList>(file_list)),
  when_to_start_(when_to_start),
  timeout_interval_(timeout),
  target_dir_(target_dir),
  is_running_(false),
  pending_xfer_count_(0)
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  file_name_ =
    Util::Config::get_task_dir() + G_DIR_SEPARATOR_S + name_.raw() + ".xml";

  logger_ = new Transfer::Logger();

  stats_ = new Transfer::Statistics();

  set_source_files(src_files);

  LOG_DD(name_
         << " got " << static_cast<int>(src_files_.size())
         << " files at startup.");

  // determine the target drive
  HalClient* hc = HalClient::instance();
  target_drive_ = hc->find_flash_by_udi(drive_udi_);

  // fire up FileWatcher
  FileWatcher* fw = FileWatcher::instance();
  fw->signal_monitor.connect(sigc::mem_fun(*this, &Task::on_monitor));
  fw->add_file(target_dir_);

  // connect to the timeout signal
  timeout_connection_ = Glib::signal_timeout().connect(
    sigc::mem_fun(*this, &Task::on_timeout),
    timeout_interval_);
}

Task::~Task()
{
  timeout_connection_.disconnect();

  delete logger_;
  delete stats_;
}

std::list< Glib::RefPtr<Gnome::Vfs::Uri> >
Task::get_source_files() const
{
  uri_list uris;
  entry_list_const_iter it(src_files_.begin());
  entry_list_const_iter end(src_files_.end());

  for (; it != end; ++it)
  {
    if ((*it)->get_inclusion() == INCLUSION_EXPLICIT)
    {
      uris.push_back((*it)->get_uri());
    }
  }

  return uris;
}

void
Task::set_name(const Glib::ustring& new_name)
{
  name_ = new_name;
  file_name_ =
    Util::Config::get_task_dir() + G_DIR_SEPARATOR_S + name_.raw() + ".xml";
}

void
Task::set_timeout(int new_interval)
{
  timeout_connection_.disconnect();
  timeout_connection_ = Glib::signal_timeout().connect(
    sigc::mem_fun(*this, &Task::on_timeout),
    new_interval);
  timeout_interval_ = new_interval;
}

void
Task::add_source_file(const Glib::RefPtr<Gnome::Vfs::Uri>& uri, Inclusion type)
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  Glib::ustring path(uri->get_path());

  if (Glib::file_test(path, Glib::FILE_TEST_IS_DIR) ||
      file_list_->filter(Glib::path_get_basename(path)) ||
      (! uri->uri_exists())) // monitor it, give it a chance
  {
    SharedFileEntry entry(new FileEntry(uri, type));

    entry->set_status(FILE_STATUS_NO_CHANGE);

    LOG_DD("adding " << path);

    src_files_.push_back(entry);
  
    FileWatcher* file_watcher = FileWatcher::instance();
    file_watcher->add_file(path);
    
    if (Glib::file_test(path, Glib::FILE_TEST_IS_DIR)) // be recursive
    {
      try
      {
	list<std::string> files(Util::File::get_files_in_dir(path));

	if (! files.empty())
	{
	  list<std::string>::iterator it(files.begin());
	  list<std::string>::iterator end(files.end());
	
	  for ( ; it != end; ++it)
	  {
	    Glib::RefPtr<Gnome::Vfs::Uri> new_uri =
	      Gnome::Vfs::Uri::create(*it);

	    add_source_file(new_uri, INCLUSION_IMPLICIT);
	  }
	}
      }
      catch (Glib::FileError& ex)
      {
	LOG_EXCEPTION_DD("Cannot open directory "
	                 << path << ": " << Glib::strerror(ex.code()));
      }
    }
  }
  else
  {
    LOG_DD("skipping " << path);
  }
}

void
Task::set_source_files(const uri_list& uris)
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  LOG_DD("setting src files");

  src_files_.clear();

  uri_list_const_iter it(uris.begin());
  uri_list_const_iter end(uris.end());

   for (; it != end; ++it)
     add_source_file(*it);

   src_files_.sort(file_entry_compare());
}

Glib::ustring
Task::get_printed_bytes_transferred() const
{
  return Util::String::get_readable_byte_count(stats_->get_bytes());
}

void
Task::start(bool mounted)
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  if (! is_running_)
  {
    LOG_DD(name_ << " starting");

    is_running_ = true;

    if (mounted)
    {
      // Add the files on the just-mounted drive to file watcher.
      FileWatcher* fw = FileWatcher::instance();

      if (data_flow_ == DATA_FLOW_REMOVABLE_TO_LOCAL)
      {
	entry_list_iter it(src_files_.begin());
	entry_list_iter end(src_files_.end());
	
	for (; it != end; ++it)
        {
          fw->add_file((*it)->get_uri()->get_path());
        }
      }
      else
      {
        fw->add_file(target_dir_);
      }
    }

    signal_started.emit(this);
  }
}

void
Task::stop(const Glib::ustring& note, bool unmounted)
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  if (is_running_)
  {
    LOG_DD(name_ << " stopped");

    is_running_ = false;

    if (unmounted)
    {
      LOG_DD("drive unmounted");

      // Remove the files on the just-unmounted drive from the file watcher.
      FileWatcher* fw = FileWatcher::instance();

      if (data_flow_ == DATA_FLOW_REMOVABLE_TO_LOCAL)
      {
	entry_list_iter it(src_files_.begin());
	entry_list_iter end(src_files_.end());
	
        for (; it != end; ++it)
        {
          LOG_DD("removing " <<  (*it)->get_uri()->get_path());

          fw->remove_file((*it)->get_uri()->get_path());
        }
      }
      else
	fw->remove_file(target_dir_);
    }

    signal_stopped.emit(this, note);
  }
}

std::tr1::shared_ptr<FileEntry>
Task::find_file_entry(const Glib::ustring& path)
{
  FileEntry* entry = 0;
  SharedFileEntry res(entry);

  list<SharedFileEntry>::iterator it(src_files_.begin());
  list<SharedFileEntry>::iterator end(src_files_.end());

  for ( ; it != end; ++it)
  {
    if ((*it)->get_uri()->get_path() == path)
    {
      res = *it;
      break;
    }
  }

  return res;
}

void
Task::on_monitor(const Glib::ustring& monitor_uri,
		 const Glib::ustring& info_uri,
		 FileStatus status)
{
  SharedFileEntry entry(find_file_entry(info_uri));

  if (info_uri != monitor_uri)
  {
    // monitor_uri is a directory where info_uri has changed.

    if ((status == FILE_STATUS_CHANGED) || (status == FILE_STATUS_DELETED))
    {
      if (entry.get()) // doesn't have to be always true, it could be ignored
      {
	entry->set_status(status);
      }
    }
    else // if (status == FILE_STATUS_CREATED)
    {
      if (! entry.get())
      {
	// Totally new file.
	if (file_list_->filter((Glib::path_get_basename(info_uri))))
	{
	  Glib::RefPtr<Gnome::Vfs::Uri> uri(Gnome::Vfs::Uri::create(info_uri));
	  add_source_file(uri, INCLUSION_IMPLICIT);
	}
      }
      else
      {
	// This is the case where eg an editor has deleted + created a file,
	// in a "Save" operation.
        entry->set_status(status);
      }
    }
  }
  else
  {
    // A single file/dir has been changed that doesn't have a parent
    // directory which is being monitored.
    // Or some URI has been created (eg drive mounted) which we were
    // explicitly monitoring as the use selected it as source file eariler.

    if (entry.get()) // this should always be true in this case
    {
      if (status == FILE_STATUS_CREATED)
      {
        // It is possible that this is a directory on a removable drive
        // which just got mounted. Previously it has probably been selected
        // as a source file at a time when it was mounted, but now the
        // program has started with the source drive unmounted, the uri
        // still being included with the additional condition in
        // add_source_file().
        // We now want to make the task object also aware of all the files
        // it contains - FileWatcher could only inform us about the top-level
        // directory.

        Glib::RefPtr<Gnome::Vfs::Uri> new_uri =
          Gnome::Vfs::Uri::create(info_uri);
        Inclusion inc = entry->get_inclusion();

        src_files_.remove(entry);
        add_source_file(new_uri, inc);
      }
      else
      {
        entry->set_status(status);
      }
    }
  }
}

gint
Task::async_transfer_callback(GnomeVFSAsyncHandle* /* handle */,
                              GnomeVFSXferProgressInfo* info,
                              gpointer callback_data)
{
  Transfer::CallbackData* data =
    static_cast<Transfer::CallbackData*>(callback_data);

  if (info->status == GNOME_VFS_XFER_PROGRESS_STATUS_OK)
  {
    if (info->phase == GNOME_VFS_XFER_PHASE_COMPLETED)
    {
        *(data->logger) << Util::i18n_printf(
          // TRANSLATORS: the first string is time
          _("%s: Transfer completed successfully. Total bytes copied: %u."),
          Util::Time::get_current_time().c_str(),
          info->total_bytes_copied) << "\n";

        data->task->get_statistics().add_bytes(info->total_bytes_copied);

        data->task->pending_xfer_count_--;

        if (data->task->pending_xfer_count_ == 0)
        {
          // Synchronization cycle completed.
          data->task->on_sync_done();
        }

        delete data;
    }

    return 1; // continue with the transfer
  }
  else if (info->status == GNOME_VFS_XFER_PROGRESS_STATUS_VFSERROR)
  {
    *(data->logger) << Util::i18n_printf(
      _("VFS error occurred: %s while copying file %s. Skipping..."),
      gnome_vfs_result_to_string(info->vfs_status),
      info->source_name) << "\n";

    return GNOME_VFS_XFER_ERROR_ACTION_SKIP;
  }
  else
  {
    return 1;
  }
}

bool
Task::on_timeout()
{
  LOG_FUNCTION_SCOPE_NORMAL_DD;

  if (! is_running_)
  {
    return true;
  }

  LOG_DD(name_ << " timeout");

  signal_begin_sync.emit(this);

  entry_list_iter it(src_files_.begin());
  entry_list_iter end(src_files_.end());

  for (; it != end; ++it)
  {
    if ((*it)->is_dir() && ((*it)->get_inclusion() == INCLUSION_EXPLICIT))
    {
      // For all directories user has chosen, create a DirTransfer with
      // files which thus implicitly included.

      Glib::ustring root_src_path((*it)->get_path());
      Transfer::DirTransfer transfer(root_src_path, target_dir_);
      entry_list_iter file_it(it);

      // Find all files within this directory and add them to transfer.
      for ( ++file_it ; file_it != end; ++file_it)
      {
	Glib::ustring source_path((*file_it)->get_path());
	Glib::ustring dest_path = transfer.build_destination_path(source_path);

	if ((! (*file_it)->is_dir()) &&
	    (*file_it)->get_uri()->uri_exists() &&
	    (source_path.find(root_src_path) !=
	     Glib::ustring::npos) && // TODO: use Uri::is_parent()
	    ((*file_it)->is_dirty() || (*file_it)->initial_check() ||
	     (! Glib::file_test(dest_path, Glib::FILE_TEST_EXISTS))))
	{
	  transfer.add_file((*file_it)->get_path());
	}
      }

      transfer.run(/*logger_, */this, &Task::async_transfer_callback);
    }
    else if ((! (*it)->is_dir()) &&
	     (*it)->get_uri()->uri_exists() &&
	     ((*it)->get_inclusion() == INCLUSION_EXPLICIT) &&
	     ((*it)->is_dirty() || (*it)->initial_check() ||
	      (! Glib::file_test(
		Glib::build_filename(target_dir_,
				     Glib::path_get_basename(
				       (*it)->get_path())),
		Glib::FILE_TEST_EXISTS))))
    {
      // Regular file.
      Glib::ustring src_path((*it)->get_path());
      Glib::ustring target_path(
	Glib::build_filename(target_dir_, Glib::path_get_basename(src_path)));

      if (! Glib::file_test(Glib::path_get_dirname(target_path),
			    Glib::FILE_TEST_EXISTS))
	Util::File::make_missing_dirs_for_file(target_path);

      bool needs_copy = false;

      if (Glib::file_test(target_path, Glib::FILE_TEST_EXISTS))
      {
	const Glib::RefPtr<Gnome::Vfs::Uri> target_uri =
	  Gnome::Vfs::Uri::create(target_path);

	if ((*it)->get_uri()->get_file_info()->get_modification_time() >
	    target_uri->get_file_info()->get_modification_time())
	{
	  needs_copy = true;
	}
      }
      else
      {
	needs_copy = true;
      }

      if (needs_copy)
      {
	try
	{
          Transfer::FileTransfer transfer(src_path, target_path);

          transfer.run(/*logger_, */this, &Task::async_transfer_callback);
	}
	catch (Gnome::Vfs::exception& ex)
	{
	  LOG_EXCEPTION_DD("gnome-vfs exception: " << ex.what());
	}
      }
    }
  }

  if (pending_xfer_count_ == 0)
  {
    signal_end_sync.emit(this);
  }

  return true;
}

void
Task::on_sync_done()
{
  stats_->inc_sync_count();
  signal_end_sync.emit(this);
}

void
Task::set_file_list(SharedFileList& new_list)
{
  if (is_running_)
  {
    stop("Updating list");
    file_list_ = new_list;
    start();
  }
  else
  {
    file_list_ = new_list;
  }
}

} // namespace TaskHandling

} // namespace USBSink
